"""MTEA-Cipher in Python, modified from MTEA-Cipher in Python to use modified
cipher and modified CBC mode.
"""

import struct
import types
import warnings

MODE_ECB = 1
MODE_CBC = 2
MODE_CFB = 3
MODE_PGP = 4
MODE_OFB = 5
MODE_CTR = 6
MODE_CCM = 8 # Unofficial
MODE_MCBC = 9 # Unofficial, modified CBC mode

def new(key, **kwargs):
    """Create an cipher object.

    Keyword arguments
    key -- the key for encrypting (and decrypting)

    Other arguments:
    mode -- Block cipher mode of operation (default MODE_ECB)
    IV -- Initialisation vector (needed with CBC/CFB)
    counter -- a callable counter (needed with CTR)
    endian -- how to use struct (default "!" (big endian/network))    """
    return MTEACipher(key, **kwargs)

################ MTEACipher class

class MTEACipher(object):
    """The cipher class

    Functions:
    encrypt -- encrypt data
    decrypt -- decrypt data

    _block -- splits the data in blocks (you may need padding)

    Constants:
    block_size = 8

    Variables:
    IV -- the initialisation vector (default None or "\x00"*8)
    counter -- counter for CTR (default None)

    """

    block_size = 8
    IV = None
    counter = None

    def __init__(self, key, **kwargs):
        """Initiate the cipher
        same arguments as for new."""
        self.key = key
        if len(key) != 16: # Check key len
            raise ValueError("Key must be 128 bit long")
        keys = kwargs.keys() # arguments
        if "mode" in keys: # check for mode
            self.mode = kwargs["mode"] # read mode
        else:
            self.mode = MODE_ECB # if not given
        if self.mode == MODE_PGP:
            raise NotImplementedError("PGP-CFB is not implemented")
        if "IV" in keys: # get iv
            self.IV = kwargs["IV"]
            if len(self.IV) != 8: # iv len = blocksize
                raise ValueError("IV must be 8 bytes long")
        elif self.mode in (MODE_CBC, MODE_MCBC, MODE_CFB): # cfb & cbc need iv
            raise ValueError("CBC, MCBC, CFB need an IV")
        elif self.mode == MODE_OFB: # ofb nonce if not given = "\x00" * 16
            self.IV = b"\x00" * 8
        if "counter" in keys: # ctr needs counter
            self.counter = kwargs["counter"]
        elif self.mode == MODE_CTR: # if ctr and counter not given
            raise ValueError("CTR needs a counter")
        if "rounds" in keys: # rounds to operate
            self.rounds = kwargs["rounds"]
        else:
            self.rounds = 32
        if "endian" in keys: # endian for struct str -> int -> str (byte order)
            self.endian = kwargs["endian"]
        else:
            self.endian = "!" # default network/big endian

    def encrypt(self, data):
        """Encrypt data.

        Keyword arguments:
        data -- the data (plaintext) to encrypt

        On OFB, encrypt and decrypt is the same.
        """

        #ECB
        if self.mode == MODE_ECB:
            if not len(data) % 8:
                out = []
                blocks=self._block(data)
                for block in blocks:
                    out.append(_encrypt(self.key, block, self.rounds // 2, self.endian))
                return b"".join(out)
            else:
                raise ValueError("Input string must be a multiple of blocksize in length")

        #CBC
        elif self.mode == MODE_CBC:
            if not len(data) % 8:
                out = [self.IV]
                blocks=self._block(data)
                for i in range(0, len(blocks)):
                    xored = xor_strings(blocks[i], out[i])
                    out.append(_encrypt(self.key, xored,self.rounds // 2, self.endian))
                return b"".join(out[1:])
            else:
                raise ValueError("Input string must be a multiple of blocksize in length")
        #MCBC
        elif self.mode == MODE_MCBC:
            if not len(data) % 8:
                out = [self.IV]
                xoreds = [b"\x00" * 8]
                blocks=self._block(data)
                for i in range(0, len(blocks)):
                    xored = xor_strings(blocks[i], out[i])
                    xoreds.append(xored)
                    tx = _encrypt(self.key,xored,self.rounds // 2, self.endian)
                    out.append(xor_strings(xoreds[i], tx))
                return b"".join(out[1:])
            else:
                raise ValueError("Input string must be a multiple of blocksize in length")
        #OFB
        elif self.mode == MODE_OFB:
            return _crypt_ofb(self.key, data, self.IV, self.rounds // 2)

        #CFB
        elif self.mode == MODE_CFB:
            if len(data) % 8 == 0:
                blocks = self._block(data)
                out = []
                fb = self.IV
                for bn in blocks:
                    tx = _encrypt(self.key, fb, self.rounds // 2, self.endian)
                    fb = xor_strings(bn, tx)
                    out.append(fb)
                return b"".join(out)
            else:
                raise ValueError("Input string must be a multiple of blocksize in length")

        #CTR
        elif self.mode == MODE_CTR:
            l = (types.IntType, types.LongType, types.FloatType) # Typelist

            blocks = self._block(data)
            out = []
            for block in blocks:
                c = self.counter()
                if type(c) in l:
                    warnings.warn(
                        "Numbers as counter-value is buggy and deprecated!",
                        DeprecationWarning)
                    n = int.from_bytes(block, 'big')
                    out.append(
                        _encrypt(
                            self.key,struct.pack(self.endian + "Q", n ^ c),
                            self.rounds // 2, self.endian))
                else:
                    n = block
                    out.append(
                        _encrypt(
                            self.key,
                            xor_strings(n, c),
                            self.rounds // 2,
                            self.endian)
                        )
            return b"".join(out)

    def decrypt(self, data):
        """Decrypt data.

        Keyword arguments:
        data -- the data (ciphertext) to decrypt.

        On OFB, encrypt and decrypt is the same.
        """
        #ECB
        if self.mode == MODE_ECB:
            if len(data) % 8 == 0:
                out = []
                blocks=self._block(data)
                for block in blocks:
                    out.append(_decrypt(self.key, block, self.rounds // 2, self.endian))
                return b"".join(out)
            else:
                raise ValueError("Input string must be a multiple of blocksize in length")

        #CBC
        elif self.mode == MODE_CBC:
            if len(data) % 8 == 0:
                out = []
                blocks = self._block(data)
                blocks = [self.IV] + blocks
                for i in range(1, len(blocks)):
                    out.append(
                        xor_strings(
                            _decrypt(
                                self.key,blocks[i],
                                self.rounds // 2,
                                self.endian),
                            blocks[i - 1])
                        )
                return b"".join(out)
            else:
                raise ValueError("Input string must be a multiple of blocksize in length")
        #MCBC
        elif self.mode == MODE_MCBC:
            if len(data) % 8 == 0:
                out = []
                txs = [b"\x00" * 8]
                blocks = self._block(data)
                blocks = [self.IV] + blocks
                for i in range(1, len(blocks)):
                    tx = _decrypt(
                        self.key,
                        xor_strings(blocks[i], txs[i - 1]),
                        self.rounds // 2,
                        self.endian
                        )
                    txs.append(tx)
                    out.append(xor_strings(tx, blocks[i - 1]))
                return b"".join(out)
            else:
                raise ValueError("Input string must be a multiple of blocksize in length")
        #OFB
        elif self.mode == MODE_OFB:
            return _crypt_ofb(self.key, data, self.IV, self.rounds // 2)

        #CFB
        elif self.mode == MODE_CFB:
            if len(data) % 8 == 0:
                blocks = self._block(data)
                out = []
                fb = self.IV
                for block in blocks:
                    tx = _encrypt(self.key, fb, self.rounds // 2, self.endian)
                    fb = block[:]
                    out.append(xor_strings(block, tx))
                return b"".join(out)
            else:
                raise ValueError("Input string must be a multiple of blocksize in length")

        #CTR
        elif self.mode == MODE_CTR:
            l = (types.IntType, types.LongType, types.FloatType)
            blocks = self._block(data)
            out = []
            for block in blocks:
                c = self.counter()
                if type(c) in l:
                    warnings.warn(
                        "Numbers as counter-value are buggy and deprecated!",
                        DeprecationWarning)
                    nc = struct.unpack(self.endian + "Q", _decrypt(self.key, block, self.rounds // 2, self.endian))
                    try:
                        out.append((nc[0] ^ c).to_bytes(8, 'big'))
                    except:
                        warnings.warn(
                            "Unable to decrypt this block, block is lost",
                            RuntimeWarning)
                        out.append(b"\x00" * 8)
                else:
                    nc = _decrypt(self.key, block, self.rounds // 2, self.endian)
                    out.append(xor_strings(nc, c))
            return b"".join(out)

    def _block(self, s):
        l = []
        rest_size = len(s) % 8
        for i in range(len(s) // 8):
            l.append(s[i * 8:((i + 1) * 8)])
        if rest_size:
            raise ValueError()
        return l

################ CBCMAC class

class CBCMAC(object):
    name = "mtea-cbcmac"
    block_size = 64
    digest_size = 8

    """Just a small implementation of the CBCMAC algorithm, based on MTEA."""
    def __init__(self, key, string=b"", endian="!"):
        warnings.warn("This is experimental!")
        self.cipher = new(key, mode=MODE_CBC, IV=b"\x00" * 8, endian=endian)
        self.text = string
        self.key = key

    @staticmethod
    def new(key, string=b"", endian="!"):
        return CBCMAC(key, string, endian)

    def update(self, string):
        self.text += string

    def copy(self):
        return CBCMAC.new(self.key, self.text, self.cipher.endian)

    def digest(self):
        return self.cipher.encrypt(self.text)[-8:]

    def hexdigest(self):
        return "%016x" % self.digest()

################ Util functions: basic encrypt/decrypt, OFB, xor
"""
This are utilities only, use them only if you know what you do.

Functions:
_crypt_ofb -- Encrypt or decrypt data in OFB mode.
_encrypt -- Encrypt one single block of data.
_decrypt -- Decrypt one single block of data.
xor_strings -- xor to strings together.
"""

def _crypt_ofb(key, data, iv=b"\x00" * 8, n=16):
    """Encrypt or decrypt data in OFB mode.

    Only use if you know what you do.

    Keyword arguments:
    key -- the key for encrypting (and decrypting)
    data -- plain / ciphertext
    iv -- initialisation vector (default b"\x00"*8)
    n -- cycles, more cycles -> more security (default 16)
    """
    def keygen(key, iv, n):
        while True:
            iv = _encrypt(key, iv, n)
            for k in iv:
                yield k
    return bytes(x ^ y for x, y in zip(data, keygen(key, iv, n)))

def _encrypt(key, block, n=16, endian="!"):
    """Encrypt one single block of data.

    Only use if you know what to do.

    Keyword arguments:
    key -- the key for encrypting (and decrypting)
    block  -- one block plaintext
    n -- cycles, one cycle is two rounds, more cycles
          -> more security and slowness (default 16)
    endian -- how struct will handle data (default "!" (big endian/network))
    """
    v0, v1 = struct.unpack(endian + "2L", block)
    k = struct.unpack(endian + "4L", key)
    sum, delta, mask = 0, 0x9e3779b9, 0xffffffff
    for round in range(n):
        sum = (sum + delta) & mask
        v0 = (v0 + (((v1 << 4) + k[0]) ^ (v1 + sum) ^ ((v1 >> 5) + k[1]))) & mask
        v1 = (v1 + (((v0 << 4) + k[2]) ^ (v0 + sum) ^ ((v0 >> 5) + k[3]))) & mask
    return struct.pack(endian + "2L", v0, v1)

def _decrypt(key, block, n=16, endian="!"):
    """Decrypt one single block of data.

    Only use if you know what to do.

    Keyword arguments:
    key -- the key for encrypting (and decrypting)
    block  -- one block ciphertext
    n -- cycles, one cycle is two rounds, more cycles =
          -> more security and slowness (default 16)
    endian -- how struct will handle data (default "!" (big endian/network))
    """
    v0,v1 = struct.unpack(endian + "2L", block)
    k = struct.unpack(endian + "4L", key)
    delta, mask = 0x9e3779b9, 0xffffffff
    sum = (delta * n) & mask
    for round in range(n):
        v1 = (v1 - (((v0 << 4) + k[2]) ^ (v0 + sum) ^ ((v0 >> 5) + k[3]))) & mask
        v0 = (v0 - (((v1 << 4) + k[0]) ^ (v1 + sum) ^ ((v1 >> 5) + k[1]))) & mask
        sum = (sum - delta) & mask
    return struct.pack(endian + "2L", v0, v1)

def xor_strings(s, t):
    """xor to strings together.

    Keyword arguments:
    s -- string one
    t -- string two
    """
    return bytes(a ^ b for a, b in zip(s, t))
