#!/usr/bin/env python3

'''Decrypts the "plain RSA" algorithm implemented by the Android version of
QQ Browser.'''

import codecs
import sys

# Public key
E, N = 65537, 245406417573740884710047745869965023463

# Prime factors of N, found via
# http://www.wolframalpha.com/input/?i=factor+245406417573740884710047745869965023463
P = 14119218591450688427
Q = 17381019776996486069

def egcd(a, b):
    # Extended Euclidian Algorithm, adapted from
    # https://en.wikibooks.org/wiki/Algorithm_Implementation/Mathematics/Extended_Euclidean_algorithm#Iterative_algorithm
    x,y, u,v = 0,1, 1,0
    while a != 0:
        q, r = b//a, b%a
        m, n = x - u*q, y - v*q
        b,a, x,y, u,v = a,r, u,v, m,n
    gcd = b
    return gcd, x, y

def find_d(p, q, e):
    phi = (p - 1) * (q - 1)
    gcd, d, _ = egcd(e, phi)
    return d

D = find_d(P, Q, E)

def decrypt(crypto, key=(D, N)):
    d, n = key
    return pow(crypto, d, n)

def decrypt_string(crypto_bytes, key=(D, N)):
    crypto = int.from_bytes(crypto_bytes, 'big')
    plain = decrypt(crypto, key)
    return plain.to_bytes(len(crypto_bytes), 'big')

def main():
    try:
        crypto_bytes = codecs.decode(sys.stdin.read().strip(), 'hex')
    except Exception:
        sys.stderr.write('%s: input must be hex\n' % sys.argv[0])
        sys.exit(1)
    plain_bytes = decrypt_string(crypto_bytes)
    sys.stdout.buffer.write(plain_bytes)

if __name__ == '__main__':
    main()
