#!/usr/bin/env python3

'''Parse nested WUP requests.'''

import struct
import sys

FLOAT = struct.Struct('f')
DOUBLE = struct.Struct('d')

def read(buf, n):
    result = buf[:n]
    if len(result) < n:
        raise ValueError('payload too small')
    return result

def read_length(mv):
    ret = read_object(mv)
    (type, tag, o), mv = ret
    if not 0 <= type <= 3 and type != 12:
        raise ValueError('expected integer for length, not type %d' % type)
    if tag != 0:
        raise ValueError('expected length tag 0, not tag %d' % tag)
    return (type, o), mv

def read_struct(mv):
    o = []
    while True:
        e, mv = read_object(mv)
        type, _, _ = e
        if type == 11:
            break
        o.append(e)
    return o, mv

def read_object(mv):
    x = int.from_bytes(read(mv, 1), 'big')
    mv = mv[1:]
    type = x & 0xf
    tag = x >> 4
    if tag == 0xf:
        tag = int.from_bytes(read(mv, 1), 'big')
        mv = mv[1:]
    if type == 0:
        o = int.from_bytes(read(mv, 1), 'big')
        mv = mv[1:]
    elif type == 1:
        o = int.from_bytes(read(mv, 2), 'big')
        mv = mv[2:]
    elif type == 2:
        o = int.from_bytes(read(mv, 4), 'big')
        mv = mv[4:]
    elif type == 3:
        o = int.from_bytes(read(mv, 8), 'big')
        mv = mv[8:]
    elif type == 4:
        o = FLOAT.unpack(read(mv, 4))[0]
        mv = mv[4:]
    elif type == 5:
        o = DOUBLE.unpack(read(mv, 8))[0]
        mv = mv[8:]
    elif type == 6:
        length = int.from_bytes(read(mv, 1), 'big')
        mv = mv[1:]
        o = read(mv, length)
        mv = mv[length:]
    elif type == 7:
        length = int.from_bytes(read(mv, 4), 'big')
        mv = mv[4:]
        o = read(mv, length)
        mv = mv[length:]
    elif type == 8: # map
        lo, mv = read_length(mv)
        ltype, length = lo
        o = []
        for _ in range(length):
            k, mv = read_object(mv)
            ktype, ktag, ko = k
            if ktag != 0:
                raise ValueError('expected key tag 0, not tag %d' % ktag)
            v, mv = read_object(mv)
            vtype, vtag, vo = v
            if vtag != 1:
                raise ValueError('expected value tag 1, not tag %d' % vtag)
            o.append(((ktype, None, ko), (vtype, None, vo)))
    elif type == 9: # list
        lo, mv = read_length(mv)
        ltype, length = lo
        o = []
        for _ in range(length):
            e, mv = read_object(mv)
            etype, etag, eo = e
            if etag != 0:
                raise ValueError('expected list tag 0, not tag %d' % etag)
            o.append((etype, None, eo))
    elif type == 10:
        o, mv = read_struct(mv)
    elif type == 11:
        if tag != 0:
            raise ValueError('expected struct ending tag 0, not tag %d' % tag)
        o = None
    elif type == 12: # zero tag
        o = 0
    elif type == 13: # byte array (simple list)
        type2 = int.from_bytes(read(mv, 1), 'big')
        mv = mv[1:]
        if type2 != 0:
            raise ValueError('expected simple list to contain type 0, not type %d' % type2)
        lo, mv = read_length(mv)
        ltype, length = lo
        buf = read(mv, length)
        mv = mv[length:]
        try:
            o = parse(buf)
        except Exception:
            o = bytes(buf)
    else:
        raise ValueError('unimplemented type %d (tag %d)' % (type, tag))
    if isinstance(o, memoryview):
        o = bytes(o)
    elem = type, tag, o
    return elem, mv

def parse(mv):
    tree = []
    while mv:
        elem, mv = read_object(mv)
        type, tag, o = elem
        if type == 11:
            raise ValueError('unexpected struct end')
        tree.append(elem)
    return tree

def print_elem(elem, indent=0):
    type, tag, o = elem
    if type == 13 and not isinstance(o, bytes):
        print(' ' * indent + '(%d, %s, (' % (type, tag))
        for elem2 in o:
            print_elem(elem2, indent + 2)
        print(' ' * indent + '))')
    elif type == 9:
        print(' ' * indent + '(%d, %s, [' % (type, tag))
        for elem2 in o:
            print_elem(elem2, indent + 2)
        print(' ' * indent + '])')
    elif type == 8:
        print(' ' * indent + '(%d, %s, {' % (type, tag))
        for k, v in o:
            print(' ' * (indent + 2) + '<')
            print_elem(k, indent + 4)
            print_elem(v, indent + 4)
            print(' ' * (indent + 2) + '>')
        print(' ' * indent + '})')
    elif type == 10:
        print(' ' * indent + '(%d, %s, {' % (type, tag))
        for elem2 in o:
            print_elem(elem2, indent + 2)
        print(' ' * indent + '})')
    else:
        print(' ' * indent + '%r' % (elem,))

def main():
    buf = sys.stdin.buffer.read()
    parsed = parse(memoryview(buf))
    for elem in parsed:
        print_elem(elem)

if __name__ == '__main__':
    main()
